from abc import ABC, abstractmethod
from typing import Optional


class DataSpec(ABC):
    @abstractmethod
    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        """Get arguments to be passed to numpyro models as a dictionary.

        Parameters
        ----------
        data:
            optional dictionary to add arguments to.

        Returns
        --------
        data:
            dictionary containing arguments.
        """
        pass



from functools import partial
from typing import List, Optional
import jax.numpy as jnp

from evofr.models.model_spec import ModelSpec
from .basis_functions import BasisFunction, Spline
from .model_factories import _renewal_model


class RenewalModel(ModelSpec):
    def __init__(
        self,
        g,
        delays,
        seed_L: Optional[int] = None,
        forecast_L: Optional[int] = None,
        k: Optional[int] = None,
        RLik=None,
        CLik=None,
        SLik=None,
        v_names: Optional[List[str]] = None,
        basis_fn: Optional[BasisFunction] = None,
    ):
        self.g_rev = jnp.flip(g, axis=-1)
        self.delays = delays
        self.seed_L = seed_L if seed_L is not None else 7
        self.forecast_L = forecast_L if forecast_L is not None else 0
        self.gen_v_names = v_names

        # Making basis expansion for Rt
        self.k = k if k else 10
        self.basis_fn = (
            basis_fn if basis_fn else Spline(s=None, order=4, k=self.k)
        )

        # Defining model likelihoods
        self.RLik = RLik
        self.CLik = CLik
        self.SLik = SLik
        self.make_model()

    def make_model(self):
        self.model_fn = partial(
            _renewal_model,
            g_rev=self.g_rev,
            delays=self.delays,
            seed_L=self.seed_L,
            forecast_L=self.forecast_L,
            RLik=self.RLik,
            CaseLik=self.CLik,
            SeqLik=self.SLik,
            gen_v_names=self.gen_v_names,
        )

    def augment_data(self, data):
        # Add feature matrix for parameterization of R
        data["X"] = self.basis_fn.make_features(data)




import jax.numpy as jnp
import numpy as np
from scipy.stats import gamma, lognorm


# https://github.com/lo-hfk/epyestim/blob/ca2ca928b744f324dade248c24a40872b69a5222/epyestim/distributions.py
def continuous_dist_to_pmf(dist):
    max_dist = dist.ppf(0.9999)
    xs = jnp.linspace(0.5, max_dist + 0.5, int(max_dist + 1))
    pmf = jnp.diff(dist.cdf(xs), prepend=0)
    return pmf / pmf.sum()


def discretise_gamma(mn, std):
    a = (mn / std) ** 2
    scale = std**2 / mn
    return continuous_dist_to_pmf(gamma(a=a, scale=scale))


def discretise_lognorm(mn, std):
    gam = 1 + std**2 / mn**2
    LN = lognorm(scale=mn / np.sqrt(gam), s=np.sqrt(np.log(gam)))
    return continuous_dist_to_pmf(LN)


def pad_delays(delays):
    lens = -jnp.array([len(delay) for delay in delays])
    lens -= min(lens)
    for i in range(len(delays)):
        delays[i] = jnp.pad(delays[i], (0, lens[i]))
    return jnp.stack(delays)  # Return as matrix


def get_standard_delays():
    gen = discretise_gamma(mn=5.2, std=1.72)
    delays = [discretise_lognorm(mn=6.9, std=2.0)]
    delays = pad_delays(delays)
    return gen, delays

def to_survivor_function(delay):
    return 1. - jnp.cumsum(delay)

def is_obs_idx(v):
    return jnp.where(jnp.isnan(v), jnp.zeros_like(v), jnp.ones_like(v))


def pad_to_obs(v, obs_idx, eps=1e-12):
    return v * obs_idx + (1 - obs_idx) * eps



import jax.numpy as jnp
import numpy as np
from jax import jit, vmap
import numpyro
import numpyro.distributions as dist
from .model_functions import forward_simulate_I, reporting_to_vec
from .model_options import GARW, NegBinomCases, DirMultinomialSeq


def _renewal_model(
    cases,
    seq_counts,
    N,
    X,
    var_names,
    g_rev,
    delays,
    seed_L,
    forecast_L,
    RLik=None,
    CaseLik=None,
    SeqLik=None,
    gen_v_names=None,
    pred=False,
):
    if RLik is None:
        RLik = GARW()
    if CaseLik is None:
        CaseLik = NegBinomCases()
    if SeqLik is None:
        SeqLik = DirMultinomialSeq()

    # If single generation time
    _g_rev = g_rev  # Assume we're using original g_rev
    if g_rev.ndim == 1:
        gmap_dim = None  # Use same generation time
    else:
        gmap_dim = 0  # Use each row
        # Specifying variant name map to column
        if gen_v_names is not None:
            v_idx = [
                gen_v_names.index(s) for s in var_names
            ]  # Match names in data to generation times
            _g_rev = _g_rev[v_idx, :]

    v_fs_I = jit(
        vmap(
            forward_simulate_I,
            in_axes=(-1, -1, gmap_dim, None, None),
            out_axes=-1,
        ),
        static_argnums=4,
    )

    T, N_variant = seq_counts.shape
    obs_range = jnp.arange(seed_L, seed_L + T, 1)

    # Computing first introduction dates
    first_obs = (np.ma.masked_invalid(np.array(seq_counts)) != 0).argmax(
        axis=0
    )
    intro_dates = np.concatenate([first_obs + d for d in np.arange(0, seed_L)])
    # intro_idx = (first_obs, np.arange(N_variant)) # Single introduction
    intro_idx = (
        intro_dates,
        np.tile(np.arange(N_variant), seed_L),
    )  # Multiple introductions

    _R = RLik.model(
        N_variant, X
    )  # likelihood on effective reproduction number

    # Add forecasted R
    if forecast_L > 0:
        R_forecast = numpyro.deterministic(
            "R_forecast", jnp.vstack((_R[-1, :],) * forecast_L)
        )
        R = jnp.vstack((_R, R_forecast))
    else:
        R = _R

    # Getting initial conditions
    intros = jnp.zeros((T + seed_L + forecast_L, N_variant))
    with numpyro.plate("N_variant", N_variant):
        logI0 = numpyro.sample("logI0", dist.Normal(0, 1)) * 3.0
        # I0 = numpyro.deterministic("I0", jnp.exp(logI0))
        I0 = jnp.exp(logI0)
        # "I0"= ~ dist.LogNormal(4.0, 5.0))
    intros = intros.at[intro_idx].set(jnp.tile(I0, seed_L))

    with numpyro.plate("rho_parms", 6):
        rho_logits = numpyro.sample("rho_logits", dist.Normal()) * 3.0

    rho = jnp.exp(jnp.append(rho_logits, 0.0))
    # rho = numpyro.deterministic("rho", rho / rho.sum())
    rho = rho / rho.sum()
    rho_vec = reporting_to_vec(rho, T)

    I_prev = jnp.clip(
        v_fs_I(intros, R, _g_rev, delays, seed_L), a_min=1e-12, a_max=1e32
    )

    # Smooth trajectory for plotting
    numpyro.deterministic(
        "I_smooth", jnp.mean(rho_vec) * jnp.take(I_prev, obs_range, axis=0)
    )

    # Compute growth rate assuming I_{t+1} = I_{t} \exp(r_{t})
    numpyro.deterministic(
        "r",
        jnp.diff(
            jnp.log(jnp.take(I_prev, obs_range, axis=0)),
            prepend=jnp.nan,
            axis=0,
        ),
    )

    # Compute expected cases
    total_prev = I_prev.sum(axis=1)
    numpyro.deterministic(
        "total_smooth_prev",
        jnp.mean(rho_vec) * jnp.take(total_prev, obs_range),
    )
    EC = numpyro.deterministic("EC", jnp.take(total_prev, obs_range) * rho_vec)

    # Evaluate case likelihood
    CaseLik.model(cases, EC, pred=pred)

    # Compute frequency
    _freq = jnp.divide(I_prev, total_prev[:, None])
    freq = numpyro.deterministic("freq", jnp.take(_freq, obs_range, axis=0))

    SeqLik.model(seq_counts, N, freq, pred)  # Evaluate frequency likelihood

    numpyro.deterministic(
        "R_ave", (_R * freq).sum(axis=1)
    )  # Getting average R

    if forecast_L > 0:
        numpyro.deterministic("freq_forecast", _freq[(seed_L + T) :, :])
        I_forecast = numpyro.deterministic(
            "I_smooth_forecast",
            jnp.mean(rho_vec) * I_prev[(seed_L + T) :, :],
        )
        numpyro.deterministic(
            "r_forecast",
            jnp.diff(jnp.log(I_forecast), prepend=jnp.nan, axis=0),
        )

    return None


def _spline_incidence_model_factory(
    CaseLik=None,
    SeqLik=None,
):
    if CaseLik is None:
        CaseLik = NegBinomCases()
    if SeqLik is None:
        SeqLik = DirMultinomialSeq()

    def _variant_model(
        cases, seq_counts, N, X, X_prime, var_names=None, pred=False
    ):
        _, N_variant = seq_counts.shape
        T, k = X.shape

        # Need some way of making the R parameter formations a bit more usable
        # Time varying base trajectory
        gam = numpyro.sample("gam", dist.HalfCauchy(0.1))

        beta_rw = numpyro.sample(
            "beta_rw", dist.GaussianRandomWalk(scale=gam, num_steps=k - 1)
        )
        beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 10.0))
        beta = numpyro.deterministic(
            "beta", beta_0 + jnp.concatenate([jnp.array([0.0]), beta_rw])
        )

        # Time varying growth advantage as random walk
        # Regularizes changes in growth advantage of variants
        gam_delta = numpyro.sample("gam_delta", dist.Exponential(rate=50))
        with numpyro.plate("N_variant_m1", N_variant - 1):
            delta_rw = numpyro.sample(
                "delta_rw",
                dist.GaussianRandomWalk(scale=gam_delta, num_steps=k),
            )

        delta = delta_rw.T
        beta_mat = beta[:, None] + jnp.hstack((delta, jnp.zeros((k, 1))))

        incidence = jnp.exp(jnp.dot(X, beta_mat))  # Variant-specific incidence
        r = numpyro.deterministic("r", X_prime @ beta_mat)

        with numpyro.plate("rho_parms", 7):
            rho = numpyro.sample("rho", dist.Beta(5.0, 5.0))
        rho_vec = reporting_to_vec(rho, T)
        numpyro.deterministic("I_smooth", jnp.mean(rho_vec) * incidence)

        # Evaluate case likelihood
        CaseLik.model(cases, rho_vec * incidence.sum(axis=1), pred=pred)

        # Compute frequency
        freq = numpyro.deterministic(
            "freq", jnp.divide(incidence, incidence.sum(axis=1)[:, None])
        )

        # Evaluate frequency likelihood
        SeqLik.model(seq_counts, N, freq, pred=pred)

        # Getting average R
        numpyro.deterministic("r_ave", (r * freq).sum(axis=1))

    return _variant_model



from functools import partial
from typing import Dict, List, Optional
import jax
import jax.numpy as jnp
import json
import numpy as np
import pandas as pd
from evofr.data import forecast_dates
from evofr.data import DataSpec
from collections import defaultdict


def get_quantile(samples: Dict, p, site):
    """Returns credible interval of size 'p' from 'samples' at 'site'.

    Parameters
    ----------
    samples:
        Dictionary with keys being site or variable names.
        Values are Arrays with shape (sample_number, site_shape).

    p:
        Percent credible interval to return.

    site:
        Name of variable to generate credible interval for.

    Returns
    -------
    Array of shape (site_shape).
    """
    q = jnp.array([0.5 * (1 - p), 0.5 * (1 + p)])
    return jnp.quantile(samples[site], q=q, axis=0)


def get_median(samples: Dict, site):
    """Returns median value across all samples for a site"""
    return jnp.median(samples[site], axis=0)


def get_quantiles(samples: Dict, ps, site):
    """Returns credible interval of sizes 'ps' from 'samples' at 'site'."""
    quants = []
    for i in range(len(ps)):
        quants.append(get_quantile(samples, ps[i], site))
    med = get_median(samples, site)
    return med, quants


def get_site_by_variant(
    samples: Dict, data: DataSpec, ps, name, site, forecast=False
):

    # Unpack variant info
    var_names = data.var_names
    dates = data.dates

    # Unpack posterior
    site_name = site + "_forecast" if forecast else site
    site = samples[site_name]
    N_variant = site.shape[-1]
    T = site.shape[-2]

    if forecast:
        dates = forecast_dates(dates, T)

    # Compute medians and hdis for ps
    site_median = jnp.median(site, axis=0)

    site_hdis = [
        jnp.quantile(site, q=jnp.array([0.5 * (1 - p), 0.5 * (1 + p)]), axis=0)
        for p in ps
    ]

    site_dict = dict()
    site_dict["date"] = []
    site_dict["location"] = []
    site_dict["variant"] = []
    site_dict[f"median_{site_name}"] = []
    for p in ps:
        site_dict[f"{site_name}_upper_{round(p * 100)}"] = []
        site_dict[f"{site_name}_lower_{round(p * 100)}"] = []

    for variant in range(N_variant):
        site_dict["date"] += list(dates)
        site_dict["location"] += [name] * T
        site_dict["variant"] += [var_names[variant]] * T
        site_dict[f"median_{site_name}"] += list(site_median[:, variant])
        for i, p in enumerate(ps):
            site_dict[f"{site_name}_upper_{round(ps[i] * 100)}"] += list(
                site_hdis[i][1, :, variant]
            )
            site_dict[f"{site_name}_lower_{round(ps[i] * 100)}"] += list(
                site_hdis[i][0, :, variant]
            )
    return site_dict


def get_freq(samples: Dict, data: DataSpec, ps, name, forecast=False):
    return get_site_by_variant(
        samples, data, ps, name, "freq", forecast=forecast
    )


def get_growth_advantage(samples, data, ps, name, rel_to=None):
    # Unpack variant info
    var_names = data.var_names
    rel_to = data.pivot if rel_to is None else rel_to

    # Get posterior samples
    ga = samples["ga"]
    ga = jnp.concatenate((ga, jnp.ones(ga.shape[0])[:, None]), axis=1)
    N_variant = ga.shape[-1]

    # Loop over ga and make relative rel_to
    for i, s in enumerate(var_names):
        if s == rel_to:
            ga = jnp.divide(ga, ga[:, i][:, None])

    # ga = jnp.divide(ga, ga[:, var_names.index(rel_to)][:, None])

    # Compute medians and quantiles
    meds = jnp.median(ga, axis=0)
    gas = []
    for i, p in enumerate(ps):
        up = 0.5 + p / 2
        lp = 0.5 - p / 2
        gas.append(jnp.quantile(ga, jnp.array([lp, up]), axis=0).T)

    # Make empty dictionary
    v_dict = dict()
    v_dict["location"] = []
    v_dict["variant"] = []
    v_dict["median_ga"] = []

    for p in ps:
        v_dict[f"ga_upper_{round(p * 100)}"] = []
        v_dict[f"ga_lower_{round(p * 100)}"] = []

    for variant in range(N_variant):
        if var_names[variant] != rel_to:
            v_dict["location"].append(name)
            v_dict["variant"].append(var_names[variant])
            v_dict["median_ga"].append(meds[variant])
            for i, p in enumerate(ps):
                v_dict[f"ga_upper_{round(p * 100)}"].append(gas[i][variant, 1])
                v_dict[f"ga_lower_{round(p * 100)}"].append(gas[i][variant, 0])

    return v_dict


class EvofrEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return round(float(obj), 3)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, jax.Array):
            return self.default(np.array(obj))
        if isinstance(obj, pd.Timestamp):
            return obj.strftime("%Y-%m-%d")
        return json.JSONEncoder.default(self, obj)


def get_sites_quantiles_json(
    samples: Dict,
    data: DataSpec,
    sites: List[str],
    ps,
    name: Optional[str] = None,
):
    export_dict = dict()

    # Save common attributes at highest level
    export_dict["ps"] = ps
    export_dict["sites"] = sites
    if name:
        export_dict["name"] = name

    # Names from dataspec
    def add_dataspec_attr(
        export_dict: dict, data, attr: str, key: Optional[str] = None
    ) -> None:
        key = key if key else attr
        if hasattr(data, attr):
            export_dict[key] = getattr(data, attr)
        return None

    add_dataspec_attr(export_dict, data, "dates", key="dates")
    add_dataspec_attr(export_dict, data, "var_names", key="variants")

    # Each site has sub-dict with its info
    for site in sites:
        site_dict = dict()
        site_samples = samples[f"{site}"]

        # Get median
        site_dict["median"] = jnp.median(site_samples, axis=0)

        # Get ps
        for i, p in enumerate(ps):
            q = jnp.array(
                [0.5 * (1 - p), 0.5 * (1 + p)]
            )  # Upper and lower bound
            site_dict[f"HDI_{round(ps[i] * 100)}"] = jnp.quantile(
                site_samples,
                q=q,
                axis=0,
            )

        # Make site dict in dict
        export_dict[site] = site_dict
    return export_dict


def get_sites_variants_json(
    samples: Dict,
    data: DataSpec,
    sites: List[str],
    ps,
    name: Optional[str] = None,
):
    export_dict = dict()

    # Save common attributes at highest level

    # Make keys for probability levels
    ps_keys = ["median"]
    for i, p in enumerate(ps):
        ps_keys.append(f"HDI_{round(ps[i] * 100)}_upper")
        ps_keys.append(f"HDI_{round(ps[i] * 100)}_lower")
    export_dict["ps"] = ps_keys

    export_dict["sites"] = sites
    if name:
        export_dict["location"] = name

    # Names from dataspec
    def add_dataspec_attr(
        export_dict: dict, data, attr: str, key: Optional[str] = None
    ) -> None:
        key = key if key else attr
        if hasattr(data, attr):
            export_dict[key] = getattr(data, attr)
        return None

    add_dataspec_attr(export_dict, data, "dates", key="dates")
    add_dataspec_attr(export_dict, data, "var_names", key="variants")
    variants = export_dict["variants"]

    # Each site has sub-dict with its info
    for site in sites:
        site_dict = dict()
        site_samples = samples[site]

        for v, variant in enumerate(variants):
            variant_dict = dict()

            # Get median
            variant_dict["median"] = jnp.median(
                site_samples[:, :, v, ...], axis=0
            )

            # Get ps
            for i, p in enumerate(ps):
                q = jnp.array(
                    [0.5 * (1 - p), 0.5 * (1 + p)]
                )  # Upper and lower bound

                # Get HDI Upper
                variant_dict[f"HDI_{round(ps[i] * 100)}_upper"] = jnp.quantile(
                    site_samples[:, :, v, ...],
                    q=q[1],
                    axis=0,
                )

                # Get HDI Lower
                variant_dict[f"HDI_{round(ps[i] * 100)}_lower"] = jnp.quantile(
                    site_samples[:, :, v, ...],
                    q=q[0],
                    axis=0,
                )
            site_dict[variant] = variant_dict

        # Make site dict in dict
        export_dict[site] = site_dict
    return export_dict


def get_sites_variants_tidy(
    samples: Dict,
    data: DataSpec,
    sites: List[str],
    dated: List[bool],
    forecasts: List[bool],
    ps,
    name: Optional[str] = None,
):
    # Save metadata
    metadata = dict()

    # Make keys for probability levels
    ps_keys = ["median"]
    for p in ps:
        ps_keys.append(f"HDI_{round(p * 100)}_upper")
        ps_keys.append(f"HDI_{round(p * 100)}_lower")
    metadata["ps"] = ps_keys

    metadata["sites"] = sites
    if name:
        metadata["location"] = [name]

    # Names from dataspec
    def add_dataspec_attr(
        export_dict: dict, data, attr: str, key: Optional[str] = None
    ) -> None:
        key = key if key else attr
        if hasattr(data, attr):
            export_dict[key] = getattr(data, attr)
        return None

    add_dataspec_attr(metadata, data, "dates", key="dates")
    add_dataspec_attr(metadata, data, "var_names", key="variants")

    # Each data entry will be tidy
    date_map = data.date_to_index

    for d, site, forecast in zip(dated, sites, forecasts):
        if forecast and d:
            # Check size of dated forecasts to generate date map
            T = samples[site].shape[1]
            forecasted_dates = forecast_dates(data.dates, T)
            forecast_date_map = {
                d: i for (i, d) in enumerate(forecasted_dates)
            }
            metadata["forecast_dates"] = forecasted_dates

            break

    def tidy_site_date(site, forecast):
        # Loop over entries of median and
        med, quants = get_quantiles(samples, ps, site)
        med, quants = np.array(med), np.array(quants)

        entries = []
        T, N_variants = med.shape

        # Are we using original or forecast dates?
        if forecast:
            _date_map = forecast_date_map
        else:
            _date_map = date_map

        for v, variant in enumerate(metadata["variants"]):
            if v + 1 > N_variants:
                continue

            for date, index in _date_map.items():
                # Make template for entries
                entry = {
                    "location": name,
                    "site": site,
                    "variant": variant,
                    "date": date.strftime("%Y-%m-%d"),
                }

                # Create median entry
                entry_med = entry.copy()
                entry_med["value"] = np.around(med[index, v], decimals=3)
                entry_med["ps"] = "median"

                # Add median entry
                entries.append(entry_med)

                # Loop over intervals of interest
                for i, p in enumerate(ps):
                    entry_lower = entry.copy()
                    entry_upper = entry.copy()

                    # Add values from intervals
                    entry_lower["value"] = np.around(
                        quants[i][0, index, v], decimals=3
                    )
                    entry_lower["ps"] = f"HDI_{round(p * 100)}_lower"

                    entry_upper["value"] = np.around(
                        quants[i][1, index, v], decimals=3
                    )
                    entry_upper["ps"] = f"HDI_{round(p * 100)}_upper"

                    # Add upper and lower bounds
                    entries.append(entry_lower)
                    entries.append(entry_upper)
        return entries

    def tidy_site_flat(site):
        # Loop over entries of median and
        med, quants = get_quantiles(samples, ps, site)
        med, quants = np.array(med), np.array(quants)

        entries = []
        N_variants = med.shape[0]

        for v, variant in enumerate(metadata["variants"]):
            if v + 1 > N_variants:
                continue

            # Make template
            entry = {
                "location": name,
                "site": site,
                "variant": variant,
            }

            # Create median entry
            entry_med = entry.copy()
            entry_med["value"] = np.around(med[v], decimals=3)
            entry_med["ps"] = "median"

            # Add median entry
            entries.append(entry_med)

            # Loop over intervals of interest
            for i, p in enumerate(ps):
                entry_lower = entry.copy()
                entry_upper = entry.copy()

                # Add values from intervals
                entry_lower["value"] = np.around(quants[i][0, v], decimals=3)
                entry_lower["ps"] = f"HDI_{round(p * 100)}_lower"

                entry_upper["value"] = np.around(quants[i][1, v], decimals=3)
                entry_upper["ps"] = f"HDI_{round(p * 100)}_upper"

                # Add upper and lower bounds
                entries.append(entry_lower)
                entries.append(entry_upper)
        return entries

    entries = []
    for d, site, forecast in zip(dated, sites, forecasts):
        tidy_site = (
            partial(tidy_site_date, forecast=forecast) if d else tidy_site_flat
        )
        entries.extend(tidy_site(site))

    tidy_dict = {"metadata": metadata, "data": entries}
    return tidy_dict


def combine_sites_tidy(tidy_dicts):
    # Combine metadata
    metadata = defaultdict(list)

    for tidy_dict in tidy_dicts:
        for key, value in tidy_dict["metadata"].items():
            metadata[key].extend([v for v in value if v not in metadata[key]])

    # Loop over data
    entries = []
    for tidy_dict in tidy_dicts:
        entries.extend(tidy_dict["data"])
    return {"metadata": metadata, "data": entries}


def save_json(out: dict, path) -> None:
    with open(path, "w") as f:
        json.dump(out, f, allow_nan=False, cls=EvofrEncoder)
    return None
